"""
OpenMM tools
============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides a number of extensions to the OpenMM simulation
toolkit.
"""

from . import bond, file, pair, reporter, system, topology, unit, utility

__all__ = ["bond", "file", "pair", "reporter", "system", "topology", "unit", "utility"]



import pathlib
import sys

import MDAnalysis as mda
from MDAnalysis.tests.datafiles import PSF_TRICLINIC, DCD_TRICLINIC
from MDAnalysis.analysis.dielectric import DielectricConstant
import numpy as np

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.analysis.electrostatics import DipoleMoment # noqa: E402

def test_class_dipole_moment():

    """
    The test cases are adapted from the "Dielectric —
    :code:`MDAnalysis.analysis.dielectric`" page from the MDAnalysis
    User Guide (https://docs.mdanalysis.org/stable/documentation_pages/analysis/dielectric.html).
    """

    universe = mda.Universe(PSF_TRICLINIC, DCD_TRICLINIC)

    diel = DielectricConstant(universe.atoms)
    diel.run()

    dm = DipoleMoment(universe.atoms).run()
    dm.calculate_relative_permittivity(300)

    pdm = DipoleMoment(universe.atoms, parallel=True).run(module="joblib",
                                                          n_jobs=1)
    pdm.calculate_relative_permittivity(300)

    # TEST CASE 1: Relative permittivity of water system
    assert np.isclose(diel.results.eps_mean, dm.results.dielectrics)
    assert np.isclose(diel.results.eps_mean, pdm.results.dielectrics)


"""
Custom OpenMM bond potentials
=============================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains implementations of commonly used bond potentials
that are not available in OpenMM, such as the finite extension nonlinear
elastic (FENE) potential. Generally, the bond potentials are named after
their LAMMPS :code:`bond_style` counterparts, if available.
"""

from typing import Union

import openmm
from openmm import unit

from .pair import wca as pwca


def _setup_bond(
    cbforce: openmm.CustomBondForce,
    global_params: dict[str, Union[float, unit.Quantity]],
    per_params: list[str],
) -> None:
    """
    Sets up a :class:`openmm.CustomBondForce` object.

    Parameters
    ----------
    cbforce : `openmm.CustomBondForce`
        Custom bond force object.

    global_params : `dict`
        Global parameters.

    per_params : `list`
        Per-particle parameters.
    """

    for param in global_params.items():
        cbforce.addGlobalParameter(*param)
    for param in per_params:
        cbforce.addPerBondParameter(param)


def fene(
    global_args: dict[str, Union[float, unit.Quantity]] = None,
    wca: bool = True,
    **kwargs,
) -> tuple[openmm.CustomBondForce, openmm.CustomNonbondedForce]:
    r"""
    Implements the finite extensible nonlinear elastic (FENE) potential
    used for bead-spring polymer models.

    The potential energy between two bonded particles is given by

    .. math::

       u_\mathrm{FENE}=-\frac{1}{2}k_{12}r_{0,12}^2
       \ln{\left[1-\left(\frac{r_{12}}{r_{0,12}}\right)^2\right]}
       +4\epsilon_{12}\left[\left(\frac{\sigma_{12}}{r_{12}}\right)^{12}
       -\left(\frac{\sigma_{12}}{r_{12}}\right)^6\right]+\epsilon_{12}

    where :math:`k_{12}` is the bond coefficient in
    :math:`\mathrm{kJ}/(\mathrm{nm}^2\cdot\mathrm{mol})`,
    :math:`r_{0,12}` is the equilibrium bond length in
    :math:`\mathrm{nm}`, :math:`\sigma_{12}` is the average particle
    size in :math:`\mathrm{nm}`, and :math:`\epsilon_{12}` is the
    dispersion energy in :math:`\mathrm{kJ/mol}`. :math:`k_{12}`,
    :math:`r_{0,12}`, :math:`\sigma_{12}` and :math:`\epsilon_{12}` are
    determined from per-bond and per-particle parameters `k`, `r0`,
    `sigma` and `epsilon`, respectively, which are set using
    :meth:`openmm.openmm.CustomBondForce.addBond` and
    :meth:`openmm.openmm.NonbondedForce.addParticle`.

    Parameters
    ----------
    global_args : `dict`, optional
        Constant values :math:`k_{12}` and :math:`r_{0,12}` to use
        instead of per-bond parameters. The corresponding per-bond
        parameters will not be registered, but the remaining
        per-bond parameters will still have to be provided in their
        default order.

    wca : `bool`, default: :code:`True`
        Determines whether the Weeks–Chandler–Andersen (WCA) potential
        is included.

    **kwargs
        Keyword arguments to be passed to
        :meth:`mdcraft.openmm.pair.wca` if :code:`wca=True`.

    Returns
    -------
    bond_fene : `openmm.CustomBondForce`
        FENE bond potential.

    pair_wca : `openmm.CustomNonbondedForce`
        WCA pair potential, if :code:`wca=True`.
    """

    bond_fene = openmm.CustomBondForce("-0.5*k*r0^2*log(1-(r/r0)^2)")
    per_args = ["k", "r0"]
    for param in global_args.keys():
        if param in per_args:
            per_args.remove(param)
    _setup_bond(bond_fene, global_args, per_args)

    if wca:
        pair_wca = pwca(**kwargs)
        return bond_fene, pair_wca

    return bond_fene



"""
Topology transformations
========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains algorithms for initializing or transforming
topologies.
"""

from typing import Any, Union
import warnings

import MDAnalysis as mda
from MDAnalysis.lib.distances import minimize_vectors
import numpy as np
from .. import FOUND_OPENMM, Q_, ureg
from .molecule import center_of_mass
from .utility import (
    is_lower_triangular,
    get_closest_factors,
    replicate,
    find_connected_nodes,
)
from .unit import strip_unit

if FOUND_OPENMM:
    from openmm import app, unit


def create_atoms(
    dimensions: Union[np.ndarray[float], "unit.Quantity", Q_, "app.Topology"],
    N: int = None,
    N_p: int = 1,
    *,
    lattice: str = None,
    length: Union[float, "unit.Quantity"] = 0.34,
    flexible: bool = False,
    bonds: bool = False,
    angles: bool = False,
    dihedrals: bool = False,
    randomize: bool = False,
    length_unit: Union["unit.Unit", ureg.Unit] = None,
    wrap: bool = False,
) -> Any:
    """
    Generates initial particle positions for coarse-grained simulations.

    Parameters
    ----------
    dimensions : `numpy.ndarray`, `openmm.unit.Quantity`, \
    `pint.Quantity`, or `openmm.app.Topology`
        System dimensions, box vectors, or lattice parameters provided
        as an array, or an OpenMM topology with system dimension 
        information.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\\mathrm{nm}`.

    N : `int`, optional
        Total number of particles :math:`N`. Must be provided for random
        melts or polymers.

    N_p : `int`, default: :code:`1`
        Number of atoms (monomers) :math:`N_\\mathrm{p}` in each
        segment (polymer chain).

        **Valid values**: :math:`1\\leq N_\\mathrm{p}\\leq N`, with
        :math:`N` divisible by :math:`N_\\mathrm{p}`.

    lattice : `str`, keyword-only, optional
        Lattice type, with the relevant length scale specified in
        `length`. If `lattice` is not specified, particle positions will
        be assigned randomly.

        .. tip::

           To build walls with the correct periodicity, set the
           :math:`z`-dimension to :code:`0` in `dimensions` and
           :code:`flexible=True`. This function will then return
           the wall particle positions and the :math:`x`- and
           :math:`y`-dimensions closest to those specified in
           `dimensions` that satisfy the lattice periodicity.

           Walls should only be built in the :math:`z`-direction.

        .. container::

           **Valid values**:

           * :code:`"fcc"`: Face-centered cubic (FCC) lattice,
             determined by the particle size :math:`a`.
           * :code:`"hcp"`: Hexagonal close-packed (HCP) lattice,
             determined by the particle size :math:`a`.
           * :code:`"cubic"`: Cubic crystal system, determined by the
             particle size :math:`a`.
           * :code:`"honeycomb"`: Honeycomb lattice (e.g., graphene),
             determined by the bond length :math:`b`.

    length : `float` or `openmm.unit.Quantity`, default: :code:`0.34`
        For random polymer positions, `length` is the bond length used in
        the random walk. For lattice systems, `length` is either the
        particle size or the bond length, depending on the lattice type.
        Has no effect if :code:`N_p=1` or :code:`lattice=None`.

        **Reference unit**: :math:`\\mathrm{nm}`.

    flexible : `bool`, default: :code:`False`
        Specifies whether `dimensions` can be modified to satisfy the
        lattice periodicity, if applicable. For example, if the provided
        :math:`z`-dimension can hold a non-integer 19.7 replicas of a
        lattice, then it is updated to reflect the width of 20 replicas.
        To ignore a direction (and make that dimension constant), such
        as when creating walls, set that dimension to :code:`0` in
        `dimensions`.

    bonds : `bool`, default: :code:`False`
        Determines whether bond information is returned for polymeric
        systems. Has no effect if :code:`N_p=1`.

    angles : `bool`, default: :code:`False`
        Determines whether angle information is returned for polymeric
        systems. Has no effect if :code:`N_p=1`.

    dihedrals : `bool`, default: :code:`False`
        Determines whether dihedral information is returned for polymeric
        systems. Has no effect if :code:`N_p=1`.

    randomize : `bool`, default: :code:`False`
        Determines whether the order of the replicated polymer positions
        are randomized. Has no effect if :code:`N_p=1`.

    length_unit : `openmm.unit.Unit` or `pint.Unit`, optional
        Length unit. If not specified, it is determined automatically
        from `dimensions` or `length`.

    wrap : `bool`, default: :code:`False`
        Determines whether particles outside the simulation box are
        wrapped back into the main unit cell.

    Returns
    -------
    positions : `numpy.ndarray` or `openmm.unit.Quantity`
        Particle positions.

        **Shape**: :math:`(N,\\,3)`.

        **Reference unit**: :math:`\\mathrm{nm}`.

    bonds : `numpy.ndarray`
        Pairs of all bonded particle indices. Only returned if
        :code:`bonds=True`.

        **Shape**: :math:`(N_\\mathrm{bonds},\\,2)`.

    angles : `numpy.ndarray`
        Triples of all particle indices that form an angle. Only
        returned if :code:`angles=True`.

        **Shape**: :math:`(N_\\mathrm{angles},\\,3)`.

    dihedrals : `numpy.ndarray`
        Quadruples of all particle indices that form a dihedral. Only
        returned if :code:`dihedrals=True`.

        **Shape**: :math:`(N_\\mathrm{dihedrals},\\,4)`.

    dimensions : `numpy.ndarray` or `openmm.unit.Quantity`
        Dimensions for lattice systems. Only returned if `lattice` is
        not :code:`None`.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\\mathrm{nm}`.
    """

    # Get raw numerical dimensions and length
    if isinstance(dimensions, app.Topology):
        pbv, length_unit = strip_unit(dimensions.getPeriodicBoxVectors(), length_unit)
        pbv = np.asarray(pbv)
    else:
        dimensions, length_unit = strip_unit(dimensions, length_unit)
        pbv = get_cell_representation(np.asarray(dimensions), "vectors")
    dimensions = np.diag(pbv)
    length, length_unit = strip_unit(length, length_unit)
    length_unit = length_unit or 1

    if lattice is None:

        # Ensure the user-specified values are valid
        if N is None:
            raise ValueError("The number of particles N must be specified.")
        if not isinstance(N, (int, np.integer)):
            raise ValueError("The number of particles N must be an integer.")
        if not (1 <= N_p <= N and isinstance(N_p, (int, np.integer))):
            emsg = (
                "The number of particles N_p in each segment must "
                "be an integer between 1 and N."
            )
            raise ValueError(emsg)
        if N_p > 1 and N % N_p != 0:
            emsg = (
                f"{N=} particles cannot be evenly divided into segments "
                f"with {N_p=} particles."
            )
            raise ValueError(emsg)

        # Generate particle positions for a random melt
        if N_p == 1:
            return (np.random.rand(N, 3) @ pbv.T) * length_unit
        else:
            topo = []

            # Determine unit cell information for each segment
            segments = N // N_p
            n_cells = get_closest_factors(segments, 3)
            cell_dims = dimensions / n_cells

            # Randomly generate a segment within the unit cell
            cell_pos = np.zeros((N_p, 3))
            cell_pos[0] = cell_dims / 4
            rng = np.random.default_rng()
            for i in range(1, N_p):
                vec = rng.random(3) * 2 - 1
                cell_pos[i] = cell_pos[i - 1] + length * vec / np.linalg.norm(vec)

            # Replicate unit cell in x-, y-, and z-directions (and
            # randomize, if desired)
            pos = replicate(cell_dims, cell_pos, n_cells)
            if randomize:
                pos = np.vstack(rng.permutation(pos.reshape((segments, -1, 3))))

            # Wrap particles past the system boundaries
            if wrap:
                for i in range(3):
                    pos[pos[:, i] < 0, i] += dimensions[i]
                    pos[pos[:, i] > dimensions[i], i] -= dimensions[i]

            topo.append((pos / dimensions @ pbv.T) * length_unit)

            # Determine all bonds
            if bonds:
                topo.append(
                    np.array(
                        [
                            (i * N_p + j, i * N_p + j + 1)
                            for i in range(segments)
                            for j in range(N_p - 1)
                        ]
                    )
                )

            # Determine all angles
            if angles:
                topo.append(
                    np.array(
                        [
                            np.arange(i * N_p + j, i * N_p + j + 3)
                            for i in range(segments)
                            for j in range(N_p - 2)
                        ]
                    )
                )

            # Determine all dihedrals
            if dihedrals:
                topo.append(
                    np.array(
                        [
                            np.arange(i * N_p + j, i * N_p + j + 4)
                            for i in range(segments)
                            for j in range(N_p - 3)
                        ]
                    )
                )

            return topo[0] if len(topo) == 1 else tuple(topo)
    else:
        around = np.around if flexible else np.floor

        # Set unit cell information
        if lattice == "cubic":
            dims = dimensions.copy()
            dims[np.isclose(dims, 0)] = 1
            n_cells = around(dims / length).astype(int)
            cell_dims = length * np.ones(3, dtype=float)
            x, y, z = (length * np.arange(n, dtype=float) for n in n_cells)
            pos = np.stack(np.meshgrid(x, y, z), axis=-1).reshape(-1, 3)
        else:
            if lattice == "fcc":
                cell_dims = length * np.array((1, np.sqrt(3), 3 * np.sqrt(6) / 3))
                cell_pos = length * np.array(
                    (
                        (0, 0, 0),
                        (0.5, np.sqrt(3) / 2, 0),
                        (0.5, np.sqrt(3) / 6, np.sqrt(6) / 3),
                        (0, 2 * np.sqrt(3) / 3, np.sqrt(6) / 3),
                        (0, np.sqrt(3) / 3, 2 * np.sqrt(6) / 3),
                        (0.5, 5 * np.sqrt(3) / 6, 2 * np.sqrt(6) / 3),
                    )
                )
            elif lattice == "hcp":
                cell_dims = length * np.array((1, np.sqrt(3), 2 * np.sqrt(6) / 3))
                cell_pos = length * np.array(
                    (
                        (0, 0, 0),
                        (0.5, np.sqrt(3) / 2, 0),
                        (0.5, np.sqrt(3) / 6, np.sqrt(6) / 3),
                        (0, 2 * np.sqrt(3) / 3, np.sqrt(6) / 3),
                    )
                )
            elif lattice == "honeycomb":
                cell_dims = length * np.array((np.sqrt(3), 3, np.inf))
                cell_pos = length * np.array(
                    (
                        (0, 0, 0),
                        (0, 1, 0),
                        (np.sqrt(3) / 2, 1.5, 0),
                        (np.sqrt(3) / 2, 2.5, 0),
                    )
                )

            # Determine unit cell multiples
            n_cells = around(dimensions / cell_dims).astype(int)
            n_cells[n_cells == 0] = 1
            cell_dims[np.isinf(cell_dims)] = 0

            # Replicate unit cell in x-, y-, and z-directions
            pos = replicate(cell_dims, cell_pos, n_cells)

        # Remove particles outside of system boundaries
        if flexible:
            n_cells[np.isclose(dimensions, 0)] = 0
            pos = pos[~np.any(pos[:, np.isclose(dimensions, 0)] > 0, axis=1)]
        else:
            pos = pos[~np.any(pos > dimensions, axis=1)]

        return (
            (
                np.divide(
                    pos, dimensions, out=np.zeros_like(pos), where=dimensions != 0
                )
                @ pbv.T
            )
            * length_unit,
            n_cells * cell_dims * length_unit,
        )


def unwrap(
    positions: np.ndarray[float],
    positions_old: np.ndarray[float],
    dimensions: np.ndarray[float],
    *,
    thresholds: float = None,
    images: np.ndarray[int] = None,
    in_place: bool = True,
) -> Union[None, tuple[np.ndarray[float], np.ndarray[float], np.ndarray[int]]]:
    r"""
    Globally unwraps particle positions.

    Parameters
    ----------
    positions : `numpy.ndarray`
        Particle positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    positions_old : `numpy.ndarray`
        Previous particle positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    dimensions : `numpy.ndarray`
        System dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    thresholds : `float`, keyword-only, optional
        Maximum distances in each direction a particle can move before
        it is considered to have crossed a boundary.

        **Reference unit**: :math:`\mathrm{nm}`.

    images : `numpy.ndarray`, keyword-only, optional
        Current image flags (or number of boundary crossings).

        **Shape**: :math:`(N,\,3)`.

    in_place : `bool`, keyword-only, default: :code:`False`
        Determines whether the input array is modified in-place.

    Returns
    -------
    positions : `numpy.ndarray`
        Unwrapped particle positions. Only returned if
        :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    positions_old : `numpy.ndarray`
        Updated previous particle positions. Only returned if
        :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    images : `numpy.ndarray`
        Updated image flags (or number of boundary crossings). Only
        returned if :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.
    """

    if thresholds is None:
        thresholds = np.min(dimensions) / 2
    if images is None:
        images = np.zeros_like(positions, dtype=int)

    dpos = positions - positions_old
    mask = np.abs(dpos) >= thresholds
    if in_place:
        images[mask] -= np.sign(dpos[mask]).astype(int)
        positions_old[:] = positions[:]
        positions += images * dimensions
    else:
        positions = positions.copy()
        positions_old = positions.copy()
        images = images.copy()
        images[mask] -= np.sign(dpos[mask]).astype(int)
        positions += images * dimensions
        return positions, positions_old, images


def unwrap_edge(
    group: mda.AtomGroup = None,
    *,
    positions: np.ndarray[float] = None,
    bonds: np.ndarray[int] = None,
    dimensions: np.ndarray[float] = None,
    thresholds: np.ndarray[float] = None,
    masses: np.ndarray[float] = None,
) -> np.ndarray[float]:
    r"""
    Locally unwraps the positions of molecules at the edge of the
    simulation box.

    Parameters
    ----------
    group : `MDAnalysis.AtomGroup`, optional
        Atom group. If not provided, the atom positions, bonds, and
        system dimensions must be provided in `positions`, `bonds`, and
        `dimensions`, respectively.

    positions : `numpy.ndarray`, keyword-only, optional
        Atom positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    bonds : `numpy.ndarray`, keyword-only, optional
        Pairs of all bonded atom indices in reference to `positions`.

        **Shape**: :math:`(N_\mathrm{bonds},\,2)`.

    dimensions : `numpy.ndarray`, keyword-only, optional
        System dimensions and, optionally, angles.

        **Shape**: :math:`(3,)` or :math:`(6,)`.

        **Reference unit**: :math:`\mathrm{nm}` (lengths) and
        :math:`^\circ` (angles).

    thresholds : `numpy.ndarray`, keyword-only, optional
        Maximum distances in each direction an atom can move before it
        is considered to have crossed a boundary.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    masses : `numpy.ndarray`, keyword-only, optional
        Atom masses. If not specified, all atoms are assumed to have the
        same mass.

        **Shape**: :math:`(N,)`.

        **Reference unit**: :math:`\mathrm{g/mol}`.

    Returns
    -------
    positions : `numpy.ndarray`
        Unwrapped atom positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.
    """

    if group is not None:
        return group.unwrap()

    elif positions is not None:
        if bonds is None:
            emsg = "Bond information must be specified in 'bonds'."
            raise ValueError(emsg)
        if dimensions is None:
            emsg = "System dimensions must be specified in 'dimensions'."
            raise ValueError(emsg)
        elif len(dimensions) == 3:
            dimensions = np.concatenate((dimensions, (90, 90, 90)))
        if thresholds is None:
            thresholds = dimensions[:3] / 2

        # Get graph of connected atoms
        bond_pairs = {}
        for a, b in bonds:
            bond_pairs.setdefault(a, []).append(b)
            bond_pairs.setdefault(b, []).append(a)

        # Determine indices of connected atoms in each molecule
        molecules = find_connected_nodes(bond_pairs)

        # Specify which atoms have had their positions updated
        done = {m[0] for m in molecules}
        todo = set(range(len(positions))).difference(done)

        # Upwrap atom positions using that of the nearest bonded atom
        # that has already been unwrapped
        while len(todo) > 0:
            for particle_index in todo:
                for bonded_index in bond_pairs[particle_index]:
                    if bonded_index in done:
                        positions[particle_index] = positions[
                            bonded_index
                        ] + minimize_vectors(
                            positions[particle_index] - positions[bonded_index],
                            dimensions,
                        )
                        done.add(particle_index)
                        break
            todo -= done

        if masses is None:
            wmsg = "No masses specified. All atoms are assumed to " "have a mass of 1."
            warnings.warn(wmsg)
            masses = np.ones(len(positions))
        elif len(masses) == len(molecules):
            masses = np.concatenate(masses)
        elif len(masses) != len(positions):
            emsg = (
                "The number of masses must be equal to the number "
                "of atoms or the number of molecules."
            )
            raise ValueError(emsg)

        # Recenter molecules using their centers of mass
        for molecule_indices in molecules:
            com = center_of_mass(
                positions=positions[molecule_indices], masses=masses[molecule_indices]
            )
            positions[molecule_indices] += (
                wrap(com, dimensions[:3], in_place=False) - com
            )

        return positions
    else:
        raise ValueError("Either 'group' or 'positions' must be specified.")


def wrap(
    positions: np.ndarray[float],
    dimensions: np.ndarray[float],
    *,
    in_place: bool = True,
) -> np.ndarray[float]:
    r"""
    Wraps particle positions back into the primary simulation cell.

    Parameters
    ----------
    positions : `numpy.ndarray`
        Particle positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    dimensions : `numpy.ndarray`
        System dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    in_place : `bool`, keyword-only, default: :code:`False`
        Determines whether the input array is modified in-place.

    Returns
    -------
    positions : `numpy.ndarray`
        Wrapped particle positions. Only returned if
        :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.
    """

    wrap_indices = (positions < 0) | (positions > dimensions)
    if in_place:
        positions[wrap_indices] -= (np.floor(positions / dimensions) * dimensions)[
            wrap_indices
        ]
    else:
        positions = positions.copy()
        positions[wrap_indices] -= (np.floor(positions / dimensions) * dimensions)[
            wrap_indices
        ]
        return positions


def reduce_box_vectors(vectors: np.ndarray[float]) -> np.ndarray[float]:
    r"""
    Performs lattice reduction on box vectors.

    Parameters
    ----------
    vectors : `numpy.ndarray`
        Box vectors :math:`\mathbf{a}`, :math:`\mathbf{b}`, and
        :math:`\mathbf{c}`, provided as rows in a matrix.

        **Shape**: :math:`(3,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    Returns
    -------
    reduced_vectors : `numpy.ndarray`
        Reduced box vectors.

        **Shape**: :math:`(3,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.
    """

    vectors = np.asarray(vectors)
    if is_lower_triangular(vectors):
        return vectors
    a, b, c = vectors
    c -= b * np.round(c[1] / b[1])
    c -= a * np.round(c[0] / a[0])
    b -= a * np.round(b[0] / a[0])
    return np.vstack((a, b, c))


def convert_cell_representation(
    *, parameters: np.ndarray[float] = None, vectors: np.ndarray[float] = None
) -> np.ndarray[float]:
    r"""
    Converts between crystallographic lattice parameters and triclinic
    box vectors.

    Parameters
    ----------
    parameters : array-like, optional
        Lattice parameters :math:`(a,\,b,\,c,\,\alpha,\,\beta,\,\gamma)`.

        **Shape**: :math:`(6,)`.

        **Reference unit**: :math:`\mathrm{nm}` (lengths) and
        :math:`^\circ` (angles).

    vectors : array-like, optional
        Box vectors :math:`\mathbf{a}`, :math:`\mathbf{b}`, and
        :math:`\mathbf{c}`, provided as rows in a matrix, in their
        reduced forms:

        .. math::

           \begin{align*}
             \mathbf{a} &= (a_x,\,0,\,0) \\
             \mathbf{b} &= (b_x,\,b_y,\,0) \\
             \mathbf{c} &= (c_x,\,c_y,\,c_z)
           \end{align*}
           \quad\mathrm{where}\quad 
           a_x>0,\,b_y>0,\,c_z>0,\,
           a_x\geq2|b_x|,\,a_x\geq2|c_x|,\,b_y\geq2|c_y|

        **Shape**: :math:`(3,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    Returns
    -------
    representation : `numpy.ndarray`
        Lattice parameters or box vectors.
    """

    if parameters is None and vectors is None:
        emsg = "Either 'parameters' or 'vectors' must be specified."
        raise ValueError(emsg)
    elif parameters is not None and vectors is not None:
        emsg = "Only one of 'parameters' or 'vectors' can be specified."
        raise ValueError(emsg)

    if parameters is not None:
        if len(parameters) == 3:
            warnings.warn("Only cell lengths are specified. Assuming cubic cell.")
            parameters = np.concatenate((parameters, (90, 90, 90)))
        elif len(parameters) != 6:
            emsg = "Invalid number of lattice parameters in 'parameters'."
            raise ValueError(emsg)
        else:
            parameters = parameters.copy()

        alpha, beta, gamma = np.radians(parameters[3:])
        vectors = np.zeros((3, 3))
        vectors[0, 0] = parameters[0]
        vectors[1, 0] = parameters[1] * np.cos(gamma)
        vectors[1, 1] = parameters[1] * np.sin(gamma)
        vectors[2, 0] = parameters[2] * np.cos(beta)
        vectors[2, 1] = (
            parameters[2]
            * (np.cos(alpha) - np.cos(beta) * np.cos(gamma))
            / np.sin(gamma)
        )
        vectors[2, 2] = np.sqrt(
            parameters[2] ** 2 - vectors[2, 0] ** 2 - vectors[2, 1] ** 2
        )
        vectors[np.isclose(vectors, 0, atol=5e-6)] = 0
        return vectors

    else:
        if not is_lower_triangular(vectors):
            vectors = reduce_box_vectors(vectors)
        parameters = np.empty(6)
        parameters[:3] = np.linalg.norm(vectors, axis=1)
        parameters[3] = np.degrees(
            np.arccos(np.dot(vectors[1], vectors[2]) / (parameters[1] * parameters[2]))
        )
        parameters[4] = np.degrees(
            np.arccos(np.dot(vectors[0], vectors[2]) / (parameters[0] * parameters[2]))
        )
        parameters[5] = np.degrees(
            np.arccos(np.dot(vectors[0], vectors[1]) / (parameters[0] * parameters[1]))
        )
        return parameters


def get_cell_representation(
    rep: np.ndarray[float], output: str, /
) -> np.ndarray[float]:
    r"""
    Gets the desired cell representation given a starting cell
    representation.

    Parameters
    ----------
    rep : `numpy.ndarray`, positional-only
        Starting cell representation. Can be system dimensions,
        lattice parameters, or box vectors.

        **Shape**: :math:`(3,)`, :math:`(6,)`, or :math:`(3,\,3)`.

    output : `str`, positional-only
        Desired cell representation format.

        **Valid values**: :code:`"dimensions"`, :code:`"parameters"`,
        and :code:`"vectors"`.

    Returns
    -------
    representation : `numpy.ndarray`
        Desired cell representation.

        **Shape**: :math:`(3,)`, :math:`(6,)`, or :math:`(3,\,3)`.
    """

    rep = np.asarray(rep)
    if rep.ndim == 1:
        if rep.shape[0] == 3:
            input_ = "dimensions"
        elif rep.shape[0] == 6:
            input_ = "parameters"
        else:
            emsg = (
                "Invalid shape for 'rep'. Dimensions or lattice "
                "parameters must be provided as a 3- or 6-element "
                "array, respectively."
            )
            raise ValueError(emsg)
    elif rep.ndim == 2:
        if rep.shape != (3, 3):
            emsg = (
                "Invalid shape for 'rep'. Box vectors must be "
                "provided as rows in a 3-by-3 matrix."
            )
            raise ValueError(emsg)
        input_ = "vectors"
    else:
        emsg = (
            "Invalid shape for 'rep'. Must be a 3- or 6-element "
            "array, or a 3-by-3 matrix."
        )
        raise ValueError(emsg)

    if input_ == output:
        return rep

    if output == "dimensions":
        if input_ == "parameters":
            return np.diag(convert_cell_representation(parameters=rep))
        elif input_ == "vectors":
            return np.diag(rep)
    elif output == "parameters":
        if input_ == "dimensions":
            return np.concatenate((rep, (90, 90, 90)))
        elif input_ == "vectors":
            return convert_cell_representation(vectors=rep)
    elif output == "vectors":
        if input_ == "dimensions":
            return np.diag(rep)
        elif input_ == "parameters":
            return convert_cell_representation(parameters=rep)



r"""
Gaussian models
===============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

The Gaussian model fits peaks, and is given by

.. math::

   y=\sum_{k=1}^na_k\exp{\left[-\left(\frac{x-b_k}{c_k}\right)^2\right]}

where :math:`a` is the amplitude, :math:`b` is the centroid (location),
:math:`c` is related to the peak width, and :math:`n` is the number of
peaks to fit.

This module provides the general Gaussian model above for any number of
terms :math:`k`, as well as convenience functions for Gaussian models
with :math:`1\leq n\leq8` analogous to MATLAB's :code:`gauss1`,
:code:`gauss2`, etc.
"""

import numpy as np


def gauss(x: np.ndarray, *args: float) -> np.ndarray:
    r"""
    General Gaussian model.

    .. math::

       y=\sum_{k=1}^na_k\exp{\left[
       -\left(\frac{x-b_k}{c_k}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        :math:`x`-values.

    *args : `float`
        Fitting parameters for the Gaussian term(s), ordered as
        :math:`a_1,\,b_1,\,c_1,\,a_2,\,b_2,\,c_2,\ldots,\,a_n,\,b_n,\,c_n`,
        where :math:`n` is the number of terms in the model. As such,
        the number of variable positional arguments must be divisible by
        :math:`3`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.

    Examples
    --------
    Generate :math:`x`- and :math:`y`-values (with error), and then use
    :func:`scipy.optimize.curve_fit` to fit coefficients for a one-term
    Gaussian model.

    >>> from scipy import optimize
    >>> rng = np.random.default_rng()
    >>> x = np.linspace(-3, 7, 10)
    >>> err = (2 * rng.random(x.shape) - 1) / 10
    >>>y=np.exp(-((x - 2) / 3) ** 2) + err
    >>> pk, _ = optimize.curve_fit(lambda x, a1, b1, c1: gauss(x, a1, b1, c1), x, y)
    >>> pk
    array([1.07262377, 1.90290018, 2.90033242])

    Evaluate the fitted :math:`y`-values using the coefficients.

    >>> gauss(x, *pk)
    array([0.06157175, 0.19415629, 0.45650326, 0.80031095, 1.04615509,
           1.01966105, 0.74103382, 0.40155281, 0.16224441, 0.04887866])
    """

    n = len(args)
    assert (
        n >= 3 and n % 3 == 0
    ), "Number of fitting parameters must be greater than and divisible by 3."
    return np.exp(-(((x[:, None] - args[1::3]) / args[2::3]) ** 2)) @ args[::3]


def gauss1(x: np.ndarray, a1: float, b1: float, c1: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss1` model from MATLAB.

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(x, a1, b1, c1)


def gauss2(
    x: np.ndarray, a1: float, b1: float, c1: float, a2: float, b2: float, c2: float
) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss2` model from MATLAB:

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}
       +a_2\exp{\left[-\left(\frac{x-b_2}{c_2}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    a2 : `float`
        Amplitude :math:`a_2` of the second Gaussian term.

    b2 : `float`
        Centroid :math:`b_2` of the second Gaussian term.

    c2 : `float`
        Peak width :math:`c_2` of the second Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(x, a1, b1, c1, a2, b2, c2)


def gauss3(
    x: np.ndarray,
    a1: float,
    b1: float,
    c1: float,
    a2: float,
    b2: float,
    c2: float,
    a3: float,
    b3: float,
    c3: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss3` model from MATLAB.

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}
       +a_2\exp{\left[-\left(\frac{x-b_2}{c_2}\right)^2\right]}
       +a_3\exp{\left[-\left(\frac{x-b_3}{c_3}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    a2 : `float`
        Amplitude :math:`a_2` of the second Gaussian term.

    b2 : `float`
        Centroid :math:`b_2` of the second Gaussian term.

    c2 : `float`
        Peak width :math:`c_2` of the second Gaussian term.

    a3 : `float`
        Amplitude :math:`a_3` of the third Gaussian term.

    b3 : `float`
        Centroid :math:`b_3` of the third Gaussian term.

    c3 : `float`
        Peak width :math:`c_3` of the third Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(x, a1, b1, c1, a2, b2, c2, a3, b3, c3)


def gauss4(
    x: np.ndarray,
    a1: float,
    b1: float,
    c1: float,
    a2: float,
    b2: float,
    c2: float,
    a3: float,
    b3: float,
    c3: float,
    a4: float,
    b4: float,
    c4: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss4` model from MATLAB.

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}
       +a_2\exp{\left[-\left(\frac{x-b_2}{c_2}\right)^2\right]}
       +a_3\exp{\left[-\left(\frac{x-b_3}{c_3}\right)^2\right]}
       +a_4\exp{\left[-\left(\frac{x-b_4}{c_4}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    a2 : `float`
        Amplitude :math:`a_2` of the second Gaussian term.

    b2 : `float`
        Centroid :math:`b_2` of the second Gaussian term.

    c2 : `float`
        Peak width :math:`c_2` of the second Gaussian term.

    a3 : `float`
        Amplitude :math:`a_3` of the third Gaussian term.

    b3 : `float`
        Centroid :math:`b_3` of the third Gaussian term.

    c3 : `float`
        Peak width :math:`c_3` of the third Gaussian term.

    a4 : `float`
        Amplitude :math:`a_4` of the fourth Gaussian term.

    b4 : `float`
        Centroid :math:`b_4` of the fourth Gaussian term.

    c4 : `float`
        Peak width :math:`c_4` of the fourth Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(x, a1, b1, c1, a2, b2, c2, a3, b3, c3, a4, b4, c4)


def gauss5(
    x: np.ndarray,
    a1: float,
    b1: float,
    c1: float,
    a2: float,
    b2: float,
    c2: float,
    a3: float,
    b3: float,
    c3: float,
    a4: float,
    b4: float,
    c4: float,
    a5: float,
    b5: float,
    c5: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss5` model from MATLAB.

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}
       +a_2\exp{\left[-\left(\frac{x-b_2}{c_2}\right)^2\right]}
       +\cdots+a_5\exp{\left[-\left(\frac{x-b_5}{c_5}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    a2 : `float`
        Amplitude :math:`a_2` of the second Gaussian term.

    b2 : `float`
        Centroid :math:`b_2` of the second Gaussian term.

    c2 : `float`
        Peak width :math:`c_2` of the second Gaussian term.

    a3 : `float`
        Amplitude :math:`a_3` of the third Gaussian term.

    b3 : `float`
        Centroid :math:`b_3` of the third Gaussian term.

    c3 : `float`
        Peak width :math:`c_3` of the third Gaussian term.

    a4 : `float`
        Amplitude :math:`a_4` of the fourth Gaussian term.

    b4 : `float`
        Centroid :math:`b_4` of the fourth Gaussian term.

    c4 : `float`
        Peak width :math:`c_4` of the fourth Gaussian term.

    a5 : `float`
        Amplitude :math:`a_5` of the fifth Gaussian term.

    b5 : `float`
        Centroid :math:`b_5` of the fifth Gaussian term.

    c5 : `float`
        Peak width :math:`c_5` of the fifth Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(x, a1, b1, c1, a2, b2, c2, a3, b3, c3, a4, b4, c4, a5, b5, c5)


def gauss6(
    x: np.ndarray,
    a1: float,
    b1: float,
    c1: float,
    a2: float,
    b2: float,
    c2: float,
    a3: float,
    b3: float,
    c3: float,
    a4: float,
    b4: float,
    c4: float,
    a5: float,
    b5: float,
    c5: float,
    a6: float,
    b6: float,
    c6: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss6` model from MATLAB.

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}
       +a_2\exp{\left[-\left(\frac{x-b_2}{c_2}\right)^2\right]}
       +\cdots+a_6\exp{\left[-\left(\frac{x-b_6}{c_6}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    a2 : `float`
        Amplitude :math:`a_2` of the second Gaussian term.

    b2 : `float`
        Centroid :math:`b_2` of the second Gaussian term.

    c2 : `float`
        Peak width :math:`c_2` of the second Gaussian term.

    a3 : `float`
        Amplitude :math:`a_3` of the third Gaussian term.

    b3 : `float`
        Centroid :math:`b_3` of the third Gaussian term.

    c3 : `float`
        Peak width :math:`c_3` of the third Gaussian term.

    a4 : `float`
        Amplitude :math:`a_4` of the fourth Gaussian term.

    b4 : `float`
        Centroid :math:`b_4` of the fourth Gaussian term.

    c4 : `float`
        Peak width :math:`c_4` of the fourth Gaussian term.

    a5 : `float`
        Amplitude :math:`a_5` of the fifth Gaussian term.

    b5 : `float`
        Centroid :math:`b_5` of the fifth Gaussian term.

    c5 : `float`
        Peak width :math:`c_5` of the fifth Gaussian term.

    a6 : `float`
        Amplitude :math:`a_6` of the sixth Gaussian term.

    b6 : `float`
        Centroid :math:`b_6` of the sixth Gaussian term.

    c6 : `float`
        Peak width :math:`c_6` of the sixth Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(
        x, a1, b1, c1, a2, b2, c2, a3, b3, c3, a4, b4, c4, a5, b5, c5, a6, b6, c6
    )


def gauss7(
    x: np.ndarray,
    a1: float,
    b1: float,
    c1: float,
    a2: float,
    b2: float,
    c2: float,
    a3: float,
    b3: float,
    c3: float,
    a4: float,
    b4: float,
    c4: float,
    a5: float,
    b5: float,
    c5: float,
    a6: float,
    b6: float,
    c6: float,
    a7: float,
    b7: float,
    c7: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss7` model from MATLAB.

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}
       +a_2\exp{\left[-\left(\frac{x-b_2}{c_2}\right)^2\right]}
       +\cdots+a_7\exp{\left[-\left(\frac{x-b_7}{c_7}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    a2 : `float`
        Amplitude :math:`a_2` of the second Gaussian term.

    b2 : `float`
        Centroid :math:`b_2` of the second Gaussian term.

    c2 : `float`
        Peak width :math:`c_2` of the second Gaussian term.

    a3 : `float`
        Amplitude :math:`a_3` of the third Gaussian term.

    b3 : `float`
        Centroid :math:`b_3` of the third Gaussian term.

    c3 : `float`
        Peak width :math:`c_3` of the third Gaussian term.

    a4 : `float`
        Amplitude :math:`a_4` of the fourth Gaussian term.

    b4 : `float`
        Centroid :math:`b_4` of the fourth Gaussian term.

    c4 : `float`
        Peak width :math:`c_4` of the fourth Gaussian term.

    a5 : `float`
        Amplitude :math:`a_5` of the fifth Gaussian term.

    b5 : `float`
        Centroid :math:`b_5` of the fifth Gaussian term.

    c5 : `float`
        Peak width :math:`c_5` of the fifth Gaussian term.

    a6 : `float`
        Amplitude :math:`a_6` of the sixth Gaussian term.

    b6 : `float`
        Centroid :math:`b_6` of the sixth Gaussian term.

    c6 : `float`
        Peak width :math:`c_6` of the sixth Gaussian term.

    a7 : `float`
        Amplitude :math:`a_7` of the seventh Gaussian term.

    b7 : `float`
        Centroid :math:`b_7` of the seventh Gaussian term.

    c7 : `float`
        Peak width :math:`c_7` of the seventh Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(
        x,
        a1,
        b1,
        c1,
        a2,
        b2,
        c2,
        a3,
        b3,
        c3,
        a4,
        b4,
        c4,
        a5,
        b5,
        c5,
        a6,
        b6,
        c6,
        a7,
        b7,
        c7,
    )


def gauss8(
    x: np.ndarray,
    a1: float,
    b1: float,
    c1: float,
    a2: float,
    b2: float,
    c2: float,
    a3: float,
    b3: float,
    c3: float,
    a4: float,
    b4: float,
    c4: float,
    a5: float,
    b5: float,
    c5: float,
    a6: float,
    b6: float,
    c6: float,
    a7: float,
    b7: float,
    c7: float,
    a8: float,
    b8: float,
    c8: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`gauss8` model from MATLAB.

    .. math::

       y=a_1\exp{\left[-\left(\frac{x-b_1}{c_1}\right)^2\right]}
       +a_2\exp{\left[-\left(\frac{x-b_2}{c_2}\right)^2\right]}
       +\cdots+a_8\exp{\left[-\left(\frac{x-b_8}{c_8}\right)^2\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a1 : `float`
        Amplitude :math:`a_1` of the first Gaussian term.

    b1 : `float`
        Centroid :math:`b_1` of the first Gaussian term.

    c1 : `float`
        Peak width :math:`c_1` of the first Gaussian term.

    a2 : `float`
        Amplitude :math:`a_2` of the second Gaussian term.

    b2 : `float`
        Centroid :math:`b_2` of the second Gaussian term.

    c2 : `float`
        Peak width :math:`c_2` of the second Gaussian term.

    a3 : `float`
        Amplitude :math:`a_3` of the third Gaussian term.

    b3 : `float`
        Centroid :math:`b_3` of the third Gaussian term.

    c3 : `float`
        Peak width :math:`c_3` of the third Gaussian term.

    a4 : `float`
        Amplitude :math:`a_4` of the fourth Gaussian term.

    b4 : `float`
        Centroid :math:`b_4` of the fourth Gaussian term.

    c4 : `float`
        Peak width :math:`c_4` of the fourth Gaussian term.

    a5 : `float`
        Amplitude :math:`a_5` of the fifth Gaussian term.

    b5 : `float`
        Centroid :math:`b_5` of the fifth Gaussian term.

    c5 : `float`
        Peak width :math:`c_5` of the fifth Gaussian term.

    a6 : `float`
        Amplitude :math:`a_6` of the sixth Gaussian term.

    b6 : `float`
        Centroid :math:`b_6` of the sixth Gaussian term.

    c6 : `float`
        Peak width :math:`c_6` of the sixth Gaussian term.

    a7 : `float`
        Amplitude :math:`a_7` of the seventh Gaussian term.

    b7 : `float`
        Centroid :math:`b_7` of the seventh Gaussian term.

    c7 : `float`
        Peak width :math:`c_7` of the seventh Gaussian term.

    a8 : `float`
        Amplitude :math:`a_8` of the eigth Gaussian term.

    b8 : `float`
        Centroid :math:`b_8` of the eigth Gaussian term.

    c8 : `float`
        Peak width :math:`c_8` of the eigth Gaussian term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return gauss(
        x,
        a1,
        b1,
        c1,
        a2,
        b2,
        c2,
        a3,
        b3,
        c3,
        a4,
        b4,
        c4,
        a5,
        b5,
        c5,
        a6,
        b6,
        c6,
        a7,
        b7,
        c7,
        a8,
        b8,
        c8,
    )



r"""
Exponential models
==================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

Exponentials are often used when the rate of change of a quantity is
proportional to the initial amount of the quantity. The general
exponential model is given by

.. math::

   y=\sum_{k=1}^na_k\exp{(b_kx)}

If the coefficient :math:`b_k` for an :math:`\exp` term is negative,
that term represents exponential decay. If the coefficient is positive,
that term represents exponential growth.

This module provides the general exponential model above for any number
of terms :math:`k`, as well as convenience functions for the one-
(:math:`k=1`) and two-term (:math:`k=2`) exponential models
analogous to MATLAB's :code:`exp1` and :code:`exp2`, respectively.

Additionally, this module has the stretched exponential function, also
known as the complementary cumulative Weibull distribution, given by

.. math::

   y=\exp{\left[-\left(\frac{x}{\alpha}\right)^\beta\right]}

where :math:`\beta` is the stretching exponent. This expression is
obtained by inserting a fractional power law into the exponential
function.

This model is generally meaningful only for :math:`x>0`. The graph of
:math:`\log{(y)}` vs. :math:`x` is characteristically stretched when
:math:`0\leq\beta\leq 1` and compressed when :math:`\beta>1` (the latter
case has less practical importance). When :math:`\beta=1`, the one-term
exponential model is recovered. When :math:`\beta=2`, the probability
density function for the normal distribution is obtained.
"""

import numpy as np


def exp(x: np.ndarray, *args: float) -> np.ndarray:
    r"""
    General exponential model.

    .. math::

       y=\sum_{k=1}^na_k\exp{(b_kx)}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    *args : `float`
        Fitting parameters for the exponential term(s), ordered as
        :math:`a_1,\,b_1,\,a_2,\,b_2,\ldots,\,a_n,\,b_n`, where
        :math:`n` is the number of terms in the model. As such, the
        number of variable positional arguments must be even.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.

    Examples
    --------
    Generate :math:`x`- and :math:`y`-values (with error), and then use
    :func:`scipy.optimize.curve_fit` to fit coefficients for a two-term
    exponential model.

    >>> from scipy import optimize
    >>> rng = np.random.default_rng()
    >>> x = np.linspace(-0.1, 0.1, 10)
    >>> err = (2 * rng.random(x.shape) - 1) / 10
    >>> y = np.exp(-8 * x) + np.exp(12 * x) + err
    >>> pk, _ = optimize.curve_fit(
            lambda x, a1, b1, a2, b2: exp(x, a1, b1, a2, b2), x, y
        )
    >>> pk
    array([ 1.13072662, -6.90042351,  0.88706719, 12.87854508])

    Evaluate the fitted :math:`y`-values using the coefficients.

    >>> exp(x, *pk)
    array([2.49915084, 2.25973234, 2.09274413, 2.00061073, 1.98962716,
           2.07080543, 2.26106343, 2.58486089, 3.07642312, 3.78274065])
    """

    n = len(args)
    if n < 2 or n % 2 != 0:
        emsg = "Number of fitting parameters must be greater than 2 and even."
        raise ValueError(emsg)
    return np.exp(args[1::2] * x[:, None]) @ args[::2]


def exp1(x: np.ndarray, a: float, b: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`exp1` model from MATLAB.

    .. math::

       y=a\exp{(bx)}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a : `float`
        Coefficient :math:`a` for the :math:`\exp` term.

    b : `float`
        Coefficient :math:`b` for the :math:`x` term in the :math:`\exp`
        term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return exp(x, a, b)


def exp2(x: np.ndarray, a: float, b: float, c: float, d: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`exp2` model from MATLAB.

    .. math::

       y=a\exp{(bx)}+c\exp{(dx)}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a : `float`
        Coefficient :math:`a` for the first :math:`\exp` term.

    b : `float`
        Coefficient :math:`b` for the :math:`x` term in the first
        :math:`\exp` term.

    c : `float`
        Coefficient :math:`a` for the second :math:`\exp` term.

    d : `float`
        Coefficient :math:`b` for the :math:`x` term in the second
        :math:`\exp` term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return exp(x, a, b, c, d)


def biexp(
    x: np.ndarray, y0: float, a: float, b: float, c: float, d: float
) -> np.ndarray:
    r"""
    Bi-exponential function.

    .. math::

       y=y_0+a\exp{\left(-\frac{x}{b}\right)}+c\exp{\left(-\frac{x}{d}\right)}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    y0 : `float`
        Offset :math:`y_0`.

    a : `float`
        Coefficient :math:`a` for the first :math:`\exp` term.

    b : `float`
        Coefficient :math:`b` for the :math:`x` term in the first
        :math:`\exp` term.

    c : `float`
        Coefficient :math:`a` for the second :math:`\exp` term.

    d : `float`
        Coefficient :math:`b` for the :math:`x` term in the second
        :math:`\exp` term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return y0 + a * np.exp(-x / b) + c * np.exp(-x / d)


def stretched_exp(x: np.ndarray, alpha: float, beta: float) -> np.ndarray:
    r"""
    Stretched exponential function.

    .. math::

       y=\exp{\left[-\left(\frac{x}{\alpha}\right)^\beta\right]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    alpha : `float`
        Scaling parameter :math:`\alpha` for :math:`x`.

    beta : `float`
        Stretching exponent :math:`\beta`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return np.exp(-((x / alpha) ** beta))



"""
Analysis base classes
=====================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains custom base classes for serial and multithreaded
data analysis with support for the multiprocessing, Dask, Joblib, and
Numba libraries for parallelization.
"""

from abc import abstractmethod
from collections.abc import Generator
import contextlib
from datetime import datetime
import logging
import multiprocessing
import os
from typing import TextIO, Union
import warnings

try:
    import dask
    from dask import distributed

    FOUND_DASK = True
except ImportError:
    FOUND_DASK = False

try:
    import joblib

    FOUND_JOBLIB = True
except ImportError:
    FOUND_JOBLIB = False

from MDAnalysis.analysis.base import AnalysisBase
from MDAnalysis.coordinates.base import ReaderBase
import numba
import numpy as np
from tqdm import tqdm


@contextlib.contextmanager
def _tqdm_joblib(tqdm_obj: tqdm) -> Generator:
    """
    A context manager for displaying a progress bar with Joblib.

    Parameters
    ----------
    tqdm_obj : `tqdm.tqdm`
        tqdm progress bar object.
    """

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs) -> None:
            tqdm_obj.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_obj
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_obj.close()


class Hash(dict):
    """
    A hash table, or an extension of the built-in `dict` with dot
    notation for accessing properties.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v
            else:
                raise TypeError("Positional arguments must be dictionaries.")
        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        if attr.beginswith("__array"):
            return super().__getattr__(attr)
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super().__delitem__(key)
        del self.__dict__[key]


class SerialAnalysisBase(AnalysisBase):
    """
    A serial analysis base object.

    Parameters
    ----------
    trajectory : `MDAnalysis.coordinates.base.ReaderBase`
        Simulation trajectory.

    verbose : `bool`, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.
    """

    def __init__(self, trajectory: ReaderBase, verbose: bool = False, **kwargs):
        super().__init__(trajectory, verbose, **kwargs)

    def run(
        self,
        start: int = None,
        stop: int = None,
        step: int = None,
        frames: Union[slice, np.ndarray[int]] = None,
        verbose: bool = None,
        **kwargs,
    ) -> "SerialAnalysisBase":
        """
        Performs the calculation in serial.

        Parameters
        ----------
        start : `int`, optional
            Starting frame for analysis.

        stop : `int`, optional
            Ending frame for analysis.

        step : `int`, optional
            Number of frames to skip between each analyzed frame.

        frames : `slice` or array-like, optional
            Index or logical array of the desired trajectory frames.

        verbose : `bool`, optional
            Determines whether detailed progress is shown.

        **kwargs
            Additional keyword arguments to pass to
            :class:`MDAnalysis.lib.log.ProgressBar`.

        Returns
        -------
        self : `SerialAnalysisBase`
            Serial analysis base object.
        """

        return super().run(start, stop, step, frames, verbose, **kwargs)

    def save(
        self,
        file: Union[str, TextIO],
        archive: bool = True,
        compress: bool = True,
        **kwargs,
    ) -> None:
        """
        Saves results to a binary or archive file in NumPy format.

        Parameters
        ----------
        file : `str` or `file`
            Filename or file-like object where the data will be saved.
            If `file` is a `str`, the :code:`.npy` or :code:`.npz`
            extension will be appended automatically if not already
            present.

        archive : `bool`, default: :code:`True`
            Determines whether the results are saved to a single archive
            file. If `True`, the data is stored in a :code:`.npz` file.
            Otherwise, the data is saved to multiple :code:`.npy` files.

        compress : `bool`, default: :code:`True`
            Determines whether the :code:`.npz` file is compressed. Has
            no effect when :code:`archive=False`.

        **kwargs
            Additional keyword arguments to pass to :func:`numpy.save`,
            :func:`numpy.savez`, or :func:`numpy.savez_compressed`,
            depending on the values of `archive` and `compress`.
        """

        if archive and compress:
            np.savez_compressed(file, **self.results, **kwargs)
        elif archive:
            np.savez(file, **self.results, **kwargs)
        else:
            for data in self.results:
                np.save(f"{file}_{data}", self.results[data], **kwargs)


class NumbaAnalysisBase(SerialAnalysisBase):
    """
    A Numba-accelerated analysis base object.

    Parameters
    ----------
    trajectory : `MDAnalysis.coordinates.base.ReaderBase`
        Simulation trajectory.

    verbose : `bool`, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.
    """

    def __init__(self, trajectory: ReaderBase, verbose: bool = False, **kwargs):
        super().__init__(trajectory, verbose, **kwargs)

    def run(
        self,
        start: int = None,
        stop: int = None,
        step: int = None,
        frames: Union[slice, np.ndarray[int]] = None,
        verbose: bool = None,
        *,
        n_threads: int = None,
        **kwargs,
    ) -> "NumbaAnalysisBase":
        """
        Performs the calculation.

        Parameters
        ----------
        start : `int`, optional
            Starting frame for analysis.

        stop : `int`, optional
            Ending frame for analysis.

        step : `int`, optional
            Number of frames to skip between each analyzed frame.

        frames : `slice` or array-like, optional
            Index or logical array of the desired trajectory frames.

        verbose : `bool`, optional
            Determines whether detailed progress is shown.

        n_threads : `int`, keyword-only, optional
            Number of threads to use for analysis.

        **kwargs
            Additional keyword arguments to pass to
            :class:`MDAnalysis.lib.log.ProgressBar`.

        Returns
        -------
        self : `NumbaAnalysisBase`
            Analysis object with results.
        """

        if n_threads is not None:
            numba.set_num_threads(n_threads)

        return super().run(
            start=start, stop=stop, step=step, frames=frames, verbose=verbose, **kwargs
        )


class ParallelAnalysisBase(SerialAnalysisBase):
    """
    A multithreaded analysis base object.

    Parameters
    ----------
    trajectory : `MDAnalysis.coordinates.base.ReaderBase`
        Simulation trajectory.

    verbose : `bool`, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.
    """

    def __init__(self, trajectory: ReaderBase, verbose: bool = False, **kwargs):
        super().__init__(trajectory, verbose, **kwargs)

    def _dask_job_block(self, indices: np.ndarray[int]) -> list:
        return [self._single_frame_parallel(i) for i in zip(indices)]

    @abstractmethod
    def _single_frame_parallel(self, index: int):
        pass

    def run(
        self,
        start: int = None,
        stop: int = None,
        step: int = None,
        frames: Union[slice, np.ndarray[int]] = None,
        verbose: bool = None,
        *,
        n_jobs: int = None,
        module: str = "multiprocessing",
        method: str = None,
        block: bool = True,
        **kwargs,
    ) -> "ParallelAnalysisBase":
        """
        Performs the calculation in parallel.

        Parameters
        ----------
        start : `int`, optional
            Starting frame for analysis.

        stop : `int`, optional
            Ending frame for analysis.

        step : `int`, optional
            Number of frames to skip between each analyzed frame.

        frames : `slice` or array-like, optional
            Index or logical array of the desired trajectory frames.

        verbose : `bool`, optional
            Determines whether detailed progress is shown.

        n_jobs : `int`, keyword-only, optional
            Number of workers. If not specified, it is automatically
            set to either the minimum number of workers required to
            fully analyze the trajectory or the maximum number of CPU
            threads available.

        module : `str`, keyword-only, default: :code:`"multiprocessing"`
            Parallelization module to use for analysis.

            **Valid values**: :code:`"dask"`, :code:`"joblib"`, and
            :code:`"multiprocessing"`.

        method : `str`, keyword-only, optional
            Specifies which Dask scheduler, Joblib backend, or
            multiprocessing start method is used.

        block : `bool`, keyword-only, default: :code:`True`
            Determines whether the trajectory is split into smaller
            blocks that are processed serially in parallel with other
            blocks. This "split–apply–combine" approach is generally
            faster since the trajectory attributes do not have to be
            packaged for each analysis run. Only available for
            :code:`module="dask"`.

        **kwargs
            Additional keyword arguments to pass to
            :func:`dask.compute`, :class:`joblib.Parallel`, or
            :class:`multiprocessing.pool.Pool`, depending on the value of
            `module`.

        Returns
        -------
        self : `ParallelAnalysisBase`
            Parallel analysis base object.
        """

        if verbose is None:
            verbose = getattr(self, "_verbose", False)
        logging.basicConfig(
            format="{asctime} | {levelname:^8s} | {message}",
            style="{",
            level=logging.INFO if verbose else logging.WARNING,
        )

        self._setup_frames(
            self._trajectory, start=start, stop=stop, step=step, frames=frames
        )
        self._prepare()

        n_jobs = min(n_jobs or np.inf, self.n_frames, len(os.sched_getaffinity(0)))
        indices = np.arange(self.n_frames)

        if verbose:
            time_start = datetime.now()

        if module == "dask" and FOUND_DASK:
            try:
                config = {"scheduler": distributed.worker.get_client(), **kwargs}
                n_jobs = min(len(config["scheduler"].get_worker_logs()), n_jobs)
            except ValueError:
                if method is None:
                    method = "processes"
                elif method not in {
                    "distributed",
                    "processes",
                    "threading",
                    "threads",
                    "single-threaded",
                    "sync",
                    "synchronous",
                }:
                    raise ValueError("Invalid Dask scheduler.")

                if method == "distributed":
                    emsg = (
                        "The Dask distributed client "
                        "(client = dask.distributed.Client(...)) "
                        "should be instantiated in the main "
                        "program (__name__ = '__main__') of "
                        "your script."
                    )
                    raise RuntimeError(emsg)
                elif method in {"threading", "threads"}:
                    emsg = (
                        "The threaded Dask scheduler is not "
                        "compatible with MDAnalysis."
                    )
                    raise ValueError(emsg)
                elif n_jobs == 1 and method not in {
                    "single-threaded",
                    "sync",
                    "synchronous",
                }:
                    method = "synchronous"
                    logging.warning(
                        f"Since {n_jobs=}, the synchronous "
                        "Dask scheduler will be used instead."
                    )
                config = {"scheduler": method} | kwargs
                if method == "processes":
                    config["num_workers"] = n_jobs

            logging.info(
                f"Starting analysis using Dask ({n_jobs=}, "
                f"scheduler={config['scheduler']})..."
            )

            jobs = []
            if block:
                for indices_ in np.array_split(indices, n_jobs):
                    jobs.append(dask.delayed(self._dask_job_block)(indices_))
            else:
                for index in indices:
                    jobs.append(dask.delayed(self._single_frame_parallel)(index))

            blocks = dask.delayed(jobs).persist(**config)
            if verbose:
                distributed.progress(blocks)
            self._results = blocks.compute(**config)
            if block:
                self._results = [r for b in self._results for r in b]

        elif module == "joblib" and FOUND_JOBLIB:
            if method is not None and method not in {
                "loky",
                "multiprocessing",
                "threading",
                None,
            }:
                raise ValueError("Invalid Joblib backend.")

            logging.info(
                "Starting analysis using Joblib " f"({n_jobs=}, backend={method})..."
            )
            with (
                _tqdm_joblib(tqdm(total=self.n_frames))
                if verbose
                else contextlib.suppress()
            ):
                self._results = joblib.Parallel(
                    n_jobs=n_jobs, backend=method, **kwargs
                )(joblib.delayed(self._single_frame_parallel)(i) for i in indices)

        else:
            if module != "multiprocessing":
                wmsg = (
                    "The Dask or Joblib library was not found, so "
                    "the native multiprocessing module will be"
                    "used instead."
                )
                warnings.warn(wmsg)

            if method is None:
                method = multiprocessing.get_start_method()
            elif method not in {"fork", "forkserver", "spawn"}:
                raise ValueError("Invalid multiprocessing start method.")

            logging.info(
                "Starting analysis using multiprocessing " f"({n_jobs=}, {method=})..."
            )
            with multiprocessing.get_context(method).Pool(n_jobs, **kwargs) as p:
                self._results = (
                    tuple(
                        tqdm(
                            p.imap(self._single_frame_parallel, indices),
                            total=self.n_frames,
                        )
                    )
                    if verbose
                    else p.map(self._single_frame_parallel, indices)
                )

        if verbose:
            logging.info(f"Analysis finished in {datetime.now() - time_start}.")

        self._conclude()
        return self


class DynamicAnalysisBase(ParallelAnalysisBase, SerialAnalysisBase):
    """
    A dynamic analysis base object.

    Parameters
    ----------
    trajectory : `MDAnalysis.coordinates.base.ReaderBase`
        Simulation trajectory.

    parallel : `bool`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.
    """

    def __init__(
        self, trajectory: ReaderBase, parallel: bool, verbose: bool = False, **kwargs
    ) -> None:

        self._parallel = parallel
        (ParallelAnalysisBase if parallel else SerialAnalysisBase).__init__(
            self, trajectory, verbose=verbose, **kwargs
        )

    def run(
        self,
        start: int = None,
        stop: int = None,
        step: int = None,
        frames: Union[slice, np.ndarray[int]] = None,
        verbose: bool = None,
        **kwargs,
    ) -> Union[SerialAnalysisBase, ParallelAnalysisBase]:
        """
        Performs the calculation.

        .. seealso::

           For parallel-specific keyword arguments, see
           :meth:`ParallelAnalysisBase.run`.

        Parameters
        ----------
        start : `int`, optional
            Starting frame for analysis.

        stop : `int`, optional
            Ending frame for analysis.

        step : `int`, optional
            Number of frames to skip between each analyzed frame.

        frames : `slice` or array-like, optional
            Index or logical array of the desired trajectory frames.

        verbose : `bool`, optional
            Determines whether detailed progress is shown.

        **kwargs
            Additional keyword arguments to pass to
            :class:`MDAnalysis.lib.log.ProgressBar`.

        Returns
        -------
        self : `SerialAnalysisBase` or `ParallelAnalysisBase`
            Analysis object with results.
        """

        return (ParallelAnalysisBase if self._parallel else SerialAnalysisBase).run(
            self,
            start=start,
            stop=stop,
            step=step,
            frames=frames,
            verbose=verbose,
            **kwargs,
        )



r"""
Power models
============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

The power model is given by

.. math::

   y=ax^b+c

This module provides the power model above and convenience functions for
the one- (:math:`c = 0`) and two-term power models in MATLAB,
:code:`power1` and :code:`power2`, respectively.
"""

import numpy as np


def power(x: np.ndarray, a: float, b: float, c: float = 0) -> np.ndarray:
    r"""
    General power model.

    .. math::

       y=ax^b+c

    Parameters
    ----------
    x : `numpy.ndarray`
        :math:`x`-values.

    a : `float`
        Coefficient for the :math:`x^b` term.

    b : `float`
        Power constant :math:`b` for the :math:`x^b` term.

    c : `float`, keyword-only, default: :code:`0`
        Constant for the :math:`y`-intercept.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return a * x**b + c


def power1(x: np.ndarray, a: float, b: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`power1` model from MATLAB.

    .. math::

       y=ax^b

    Parameters
    ----------
    x : `numpy.ndarray`
        :math:`x`-values.

    a : `float`
        Coefficient for the :math:`x^b` term.

    b : `float`
        Power constant :math:`b` for the :math:`x^b` term.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return power(x, a, b)


def power2(x: np.ndarray, a: float, b: float, c: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`power2` model from MATLAB.

    .. math::

       y=ax^b+c

    Parameters
    ----------
    x : `numpy.ndarray`
        :math:`x`-values.

    a : `float`
        Coefficient for the :math:`x^b` term.

    b : `float`
        Power constant :math:`b` for the :math:`x^b` term.

    c : `float`
        Constant for the :math:`y`-intercept.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return power(x, a, b, c)



import pathlib
import sys

import MDAnalysis as mda
from MDAnalysis.tests.datafiles import DCD, PSF
from MDAnalysis.analysis.base import AnalysisFromFunction
import numpy as np

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.analysis.polymer import Gyradius # noqa: E402

rng = np.random.default_rng()
universe = mda.Universe(PSF, DCD)
protein = universe.select_atoms("protein")

def test_class_gyradius():

    """
    The reference implementation is adapted from the "Writing your own
    trajectory analysis" section of the MDAnalysis User Guide
    (https://userguide.mdanalysis.org/stable/examples/analysis/custom_trajectory_analysis.html).
    """

    def radius_of_gyration(group):
        positions = group.positions
        masses = group.masses
        center_of_mass = group.center_of_mass()
        r_sq = (positions - center_of_mass) ** 2
        r_ssq = np.array((r_sq.sum(axis=1),
                          (r_sq[:, [1, 2]]).sum(axis=1),
                          (r_sq[:, [0, 2]]).sum(axis=1),
                          (r_sq[:, [0, 1]]).sum(axis=1)))
        return np.sqrt((masses * r_ssq).sum(axis=1) / masses.sum())

    rog = AnalysisFromFunction(radius_of_gyration, universe.trajectory,
                               protein).run()

    # TEST CASE 1: Time series of overall radii of gyration
    gyr = Gyradius(protein, grouping="residues").run()
    pgyr = Gyradius(
        protein,
        grouping="residues",
        parallel=True
    ).run(module="joblib", n_jobs=1)
    assert np.allclose(rog.results["timeseries"][:, 0], gyr.results.gyradii[0])
    assert np.allclose(rog.results["timeseries"][:, 0], pgyr.results.gyradii[0])

    # TEST CASE 2: Time series of radius of gyration components
    gyr = Gyradius(protein, grouping="residues", components=True).run()
    pgyr = Gyradius(
        protein,
        grouping="residues",
        components=True,
        parallel=True
    ).run(module="joblib", n_jobs=1)
    assert np.allclose(rog.results["timeseries"][:, 1:],
                       gyr.results.gyradii[0])
    assert np.allclose(rog.results["timeseries"][:, 1:],
                       pgyr.results.gyradii[0])


"""
Simulation state data and trajectory analysis
=============================================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides a variety of classes for analyzing simulation
trajectories.
"""

from . import (
    base,
    electrostatics,
    polymer,
    profile,
    reader,
    structure,
    thermodynamics,
    transport,
)

__all__ = [
    "base",
    "electrostatics",
    "polymer",
    "profile",
    "reader",
    "structure",
    "thermodynamics",
    "transport",
]



import pathlib
import sys

import MDAnalysis as mda
from MDAnalysis.tests.datafiles import waterDCD, waterPSF
from MDAnalysis.analysis.lineardensity import LinearDensity
import numpy as np
import pytest

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.analysis.profile import DensityProfile # noqa: E402

universe = mda.Universe(waterPSF, waterDCD)

def test_class_density_profile():

    """
    The first two test cases are adapted from the "Computing mass and
    charge density on each axis" page from the MDAnalysis User Guide
    (https://userguide.mdanalysis.org/stable/examples/analysis/volumetric/linear_density.html).
    """

    ld = LinearDensity(universe.atoms, "residues").run()
    dp = DensityProfile(universe.atoms, "residues", axes="xy", n_bins=200,
                        average=False).run()
    pdp = DensityProfile(universe.atoms, "residues", axes="xy", n_bins=200,
                         parallel=True).run(module="joblib", n_jobs=1)

    for ax in "xy":
        number_density = (0.602214076 * ld.results[ax].mass_density
                          / universe.residues.masses[0])
        charge_density = 0.602214076 * ld.results[ax].charge_density

        # TEST CASE 1: Number density profiles
        assert np.allclose(number_density,
                           dp.results.number_densities[ax].mean(axis=1))
        assert np.allclose(number_density, pdp.results.number_densities[ax])

        # TEST CASE 2: Charge density profiles
        assert np.allclose(charge_density,
                           dp.results.charge_densities[ax].mean(axis=0))
        assert np.allclose(charge_density, pdp.results.charge_densities[ax])

    # TEST CASE 3: Wrong number of dielectric constants
    with pytest.raises(ValueError):
        dp.calculate_potential_profiles(dielectrics=(78, 78, 78))

    # TEST CASE 4: Invalid axes
    for axes in ["a", 0, [0, 1]]:
        with pytest.raises(ValueError):
            dp.calculate_potential_profiles(axes, 78)

    # TEST CASE 5: Invalid or wrong number of surface charge densities
    tests = [(0, 0, 0), np.zeros((3, 9))]
    for sigmas_q in tests:
        with pytest.raises(ValueError):
            dp.calculate_potential_profiles(dielectrics=78, sigmas_q=sigmas_q)

    # TEST CASE 6: Invalid or wrong number of potential differences
    for dVs in tests:
        with pytest.raises(ValueError):
            dp.calculate_potential_profiles(dielectrics=78, dVs=dVs)

    # TEST CASE 7: Invalid or wrong number of thresholds
    for thresholds in tests:
        with pytest.raises(ValueError):
            dp.calculate_potential_profiles(dielectrics=78,
                                            thresholds=thresholds)

    # TEST CASE 8: Invalid or wrong number of left boundary potentials
    for V0s in tests:
        with pytest.raises(ValueError):
            dp.calculate_potential_profiles(dielectrics=78, V0s=V0s)

    # TEST CASE 9: Wrong number of methods
    with pytest.raises(ValueError):
        dp.calculate_potential_profiles(
            dielectrics=78,
            methods=("integral", "integral", "matrix")
        )

    # TEST CASE 10: Wrong number of booleans for 'pbcs'
    with pytest.raises(ValueError):
        dp.calculate_potential_profiles(dielectrics=78,
                                        pbcs=(True, True, False))

    # TEST CASE 11: Invalid axis
    with pytest.raises(ValueError):
        dp.calculate_potential_profiles("z", 78)

    # TEST CASE 12: Potential profiles from integration
    dp.calculate_potential_profiles(dielectrics=78, sigmas_q=0,
                                    methods="integral")
    for ax in "xy":
        assert np.allclose(dp.results.potentials[ax], 0)

    # TEST CASE 13: Potential profiles from system of equations
    dp.calculate_potential_profiles(dielectrics=78, sigmas_q=0,
                                    methods="matrix")
    for ax in "xy":
        assert np.allclose(dp.results.potentials[ax], 0)


r"""
Polynomial models
=================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

Polynomial models for curves are given by

.. math::

   y=\sum_{k=0}^np_kx^k

where :math:`n+1` is the order of the polynomial and :math:`n` is the
degree of the polynomial. The order gives the number of coefficients to
be fit, and the degree gives the highest power of the predictor variable.

Polynomials are often used when a simple empirical model is required.
You can use the polynomial model for interpolation or extrapolation, or
to characterize data using a global fit.

This module provides the general polynomial model above for any integer
:math:`n\geq0`, as well as convenience functions for polynomial models
with :math:`1\leq n \leq9` analogous to MATLAB's :code:`poly1`,
:code:`poly2`, etc.

Note that in MATLAB, the polynomial models have the form

.. math::

   y=\sum_{i=1}^{n+1}p_ix^{n+1-i}
"""

import numpy as np


def poly(x: np.ndarray, *args: float) -> np.ndarray:
    r"""
    General polynomial model.

    .. math::

       y=\sum_{k=0}^np_kx^k

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    *args : `float`
        Fitting parameters for each :math:`x^k` term, starting with the
        :math:`x^0` term. The number of positional arguments,
        :math:`n+1`, determines the order :math:`n` of the polynomial
        model.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.

    Examples
    --------
    Generate :math:`x`- and :math:`y`-values (with error), and then use
    :meth:`scipy.optimize.curve_fit` to fit coefficients for a third-order
    polynomial.

    >>> from scipy import optimize
    >>> rng = np.random.default_rng()
    >>> x = np.arange(-2, 3)
    >>> err = (2 * rng.random(x.shape) - 1) / 10
    >>> y = x ** 2+2 * x+1+err
    >>> pk, _ = optimize.curve_fit(lambda x, p0, p1, p2: poly(x, p0, p1, p2), x, y)
    >>> pk
    array([0.98225396, 2.0243695 , 1.01370609])

    Evaluate the fitted :math:`y`-values using the coefficients.

    >>> poly(x, *pk)
    array([ 0.9883393 , -0.02840946,  0.98225396,  4.02032955,  9.08581731])
    """

    return args @ (x ** np.arange(len(args))[:, None])


def poly1(x: np.ndarray, p1: float, p2: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly1` model from MATLAB:

    .. math::

       y=p_1x+p_2

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x` term.

    p2 : `float`
        Constant term :math:`p_2`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p2, p1)


def poly2(x: np.ndarray, p1: float, p2: float, p3: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly2` model from MATLAB:

    .. math::

       y=p_1x^2+p_2x+p_3

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^2` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x` term.

    p3 : `float`
        Constant term :math:`p_3`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p3, p2, p1)


def poly3(x: np.ndarray, p1: float, p2: float, p3: float, p4: float) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly3` model from MATLAB:

    .. math::

       y=p_1x^3+p_2x^2+p_3x+p_4

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^3` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x^2` term.

    p3 : `float`
        Coefficient :math:`p_3` for the :math:`x` term.

    p4 : `float`
        Constant term :math:`p_4`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p4, p3, p2, p1)


def poly4(
    x: np.ndarray, p1: float, p2: float, p3: float, p4: float, p5: float
) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly4` model from MATLAB:

    .. math::

       y=p_1x^4+p_2x^3+\cdots+p_5

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^4` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x^3` term.

    p3 : `float`
        Coefficient :math:`p_3` for the :math:`x^2` term.

    p4 : `float`
        Coefficient :math:`p_4` for the :math:`x` term.

    p5 : `float`
        Constant term :math:`p_5`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p5, p4, p3, p2, p1)


def poly5(
    x: np.ndarray, p1: float, p2: float, p3: float, p4: float, p5: float, p6: float
) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly5` model from MATLAB:

    .. math::

       y=p_1x^5+p_2x^4+\cdots+p_6

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^5` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x^4` term.

    p3 : `float`
        Coefficient :math:`p_3` for the :math:`x^3` term.

    p4 : `float`
        Coefficient :math:`p_4` for the :math:`x^2` term.

    p5 : `float`
        Coefficient :math:`p_5` for the :math:`x` term.

    p6 : `float`
        Constant term :math:`p_6`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p6, p5, p4, p3, p2, p1)


def poly6(
    x: np.ndarray,
    p1: float,
    p2: float,
    p3: float,
    p4: float,
    p5: float,
    p6: float,
    p7: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly6` model from MATLAB:

    .. math::

       y=p_1x^6+p_2x^5+\cdots+p_7

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^6` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x^5` term.

    p3 : `float`
        Coefficient :math:`p_3` for the :math:`x^4` term.

    p4 : `float`
        Coefficient :math:`p_4` for the :math:`x^3` term.

    p5 : `float`
        Coefficient :math:`p_5` for the :math:`x^2` term.

    p6 : `float`
        Coefficient :math:`p_6` for the :math:`x` term.

    p7 : `float`
        Constant term :math:`p_7`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p7, p6, p5, p4, p3, p2, p1)


def poly7(
    x: np.ndarray,
    p1: float,
    p2: float,
    p3: float,
    p4: float,
    p5: float,
    p6: float,
    p7: float,
    p8: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly7` model from MATLAB:

    .. math::

       y=p_1x^7+p_2x^6+\cdots+p_8

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^7` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x^6` term.

    p3 : `float`
        Coefficient :math:`p_3` for the :math:`x^5` term.

    p4 : `float`
        Coefficient :math:`p_4` for the :math:`x^4` term.

    p5 : `float`
        Coefficient :math:`p_5` for the :math:`x^3` term.

    p6 : `float`
        Coefficient :math:`p_6` for the :math:`x^2` term.

    p7 : `float`
        Coefficient :math:`p_7` for the :math:`x` term.

    p8 : `float`
        Constant term :math:`p_8`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p8, p7, p6, p5, p4, p3, p2, p1)


def poly8(
    x: np.ndarray,
    p1: float,
    p2: float,
    p3: float,
    p4: float,
    p5: float,
    p6: float,
    p7: float,
    p8: float,
    p9: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly8` model from MATLAB:

    .. math::

       y=p_1x^8+p_2x^7+\cdots+p_9

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^8` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x^7` term.

    p3 : `float`
        Coefficient :math:`p_3` for the :math:`x^6` term.

    p4 : `float`
        Coefficient :math:`p_4` for the :math:`x^5` term.

    p5 : `float`
        Coefficient :math:`p_5` for the :math:`x^4` term.

    p6 : `float`
        Coefficient :math:`p_6` for the :math:`x^3` term.

    p7 : `float`
        Coefficient :math:`p_7` for the :math:`x^2` term.

    p8 : `float`
        Coefficient :math:`p_8` for the :math:`x` term.

    p9 : `float`
        Constant term :math:`p_9`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p9, p8, p7, p6, p5, p4, p3, p2, p1)


def poly9(
    x: np.ndarray,
    p1: float,
    p2: float,
    p3: float,
    p4: float,
    p5: float,
    p6: float,
    p7: float,
    p8: float,
    p9: float,
    p10: float,
) -> np.ndarray:
    r"""
    Convenience function for the :code:`poly9` model from MATLAB:

    .. math::

       y=p_1x^9+p_2x^8+\cdots+p_{10}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    p1 : `float`
        Coefficient :math:`p_1` for the :math:`x^9` term.

    p2 : `float`
        Coefficient :math:`p_2` for the :math:`x^8` term.

    p3 : `float`
        Coefficient :math:`p_3` for the :math:`x^7` term.

    p4 : `float`
        Coefficient :math:`p_4` for the :math:`x^6` term.

    p5 : `float`
        Coefficient :math:`p_5` for the :math:`x^5` term.

    p6 : `float`
        Coefficient :math:`p_6` for the :math:`x^4` term.

    p7 : `float`
        Coefficient :math:`p_7` for the :math:`x^3` term.

    p8 : `float`
        Coefficient :math:`p_9` for the :math:`x^2` term.

    p9 : `float`
        Coefficient :math:`p_9` for the :math:`x` term.

    p10 : `float`
        Constant term :math:`p_{10}`.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.
    """

    return poly(x, p10, p9, p8, p7, p6, p5, p4, p3, p2, p1)
